/*------------------------------------------------------------+
 | striphr.c -- 5/15/1992 public domain                       |
 |             by Alex Matulich, Unicorn Research Corporation |
 | Strip hard carriage returns from a text file in such a way |
 | that the paragraph formatting is not lost.  A newline char |
 | will be kept in the file if:                               |
 |       (1)   It is followed by whitespace, or               |
 |       (2)   It is preceded by a newline.                   |
 | For this program to work, the EOL code in the text must be |
 | with the OS being run.  Use StripCR or AddCR as necessary. |
 +------------------------------------------------------------*/
                                                
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#define OUTBUFSIZ 16384

unsigned short replace, bufptr = 0;
signed char *buf = NULL;
FILE *infile, *outfile;
long inpos = 0L, outpos = 0L;       /* file position trackers */

int writebuf(void);


void main(int argc, char *argv[])
{
short err = 0,
   plast, ptr = 1, pnext;  /* pointers to characters in lookahead buffer */
signed char lk[3];         /* lookahead buffer */

if (argc != 3 && argc != 2) {
   fputs("StripHR by Unicorn Research Corporation\nIntelligently strip unwanted hard returns from text files\nwhile retaining paragraph formatting.\n\nUsage:  StripHR infile [outfile]\n\n", stdout);
   fputs("If outfile is omitted, then infile will be replaced.\nA hard return is kept if it is followed by any white space (e.g.\ntabs, spaces, returns) or if preceded by another hard return.\n", stdout);
   fputs("The end-of-line code in the text file must be compatible with your\noperating system -- use StripCR / AddCR as necessary.\n", stdout);
   return;
   }
if ((infile = fopen( argv[1], (replace = (argc == 2)) ? "r+" : "r" )) == NULL) {
   fputs("Could not open input file.\n", stdout);
   return;
   }
if (replace)
   outfile = infile;
else if ((outfile = fopen( argv[2], "w" )) == NULL) {
   fputs("Could not open output file.\n", stdout);
   fclose(infile);
   return;
   }
if ((buf = (signed char *)malloc(OUTBUFSIZ)) == NULL) {
   fputs("Not enough memory!\n", stdout);
   goto cleanup;
   }

for (inpos = 0; inpos < 3; inpos++)  /* read first three characters */
   if ((lk[inpos] = fgetc(infile)) == EOF)
      { err = 2;  goto clean2; }
buf[bufptr++] = lk[0];               /* record first char in output buffer */

do {  /* process the file */
   plast = (ptr + 2) % 3;
   pnext = (ptr + 1) % 3;
   if (lk[ptr] != '\n'                                /* if not newline */
       || (lk[ptr] == '\n' &&                         /* or if newline and */
       (isspace(lk[pnext]) || lk[plast] == '\n'))) {  /* whitespace, then */
      buf[bufptr++] = lk[ptr];                        /* keep the character */
      if (replace) inpos = ftell(infile);
      if (bufptr == OUTBUFSIZ) if (err = writebuf()) break;
      }
   else {  /* otherwise substitute a space for the newline */
      buf[bufptr++] = ' ';
      if (replace) inpos = ftell(infile);
      if (bufptr == OUTBUFSIZ) if (err = writebuf()) break;
      }
   ptr = pnext;
   } while ((lk[plast] = fgetc(infile)) != EOF);

buf[bufptr++] = lk[pnext];          /* write out last character */
if (!err && bufptr) err = writebuf();

clean2:
if (err == 1)
   fputs("Error writing output file.\n", stdout);
else if (err == 2)
   fputs("Error reading input file.\n", stdout);
else if (replace) {
   fseek(outfile, outpos, 0);
   fputc('\x1A', outfile);  /* output a ctrl-Z */
   fputs(argv[1], stdout);
   fputs(" is now stripped of offending hard returns.\nDelete all text past the ctrl-Z character.\n", stdout);
   }
else {
   fputs("HR-Stripped file ", stdout);
   fputs(argv[2], stdout);
   fputs(" successfully created.\n", stdout);
   }

cleanup:
free(buf);
fclose(infile);
if (!replace) fclose(outfile);
}


/* write the buffer to the output file, when full or when done */
int writebuf()
{
unsigned short i = 0;
if (replace) if (fseek(outfile, outpos, 0)) return 1;
while (i < bufptr)
   if (fputc(buf[i++], outfile) == EOF) return 1;
if (replace) {
   outpos = ftell(outfile);
   if (fseek(infile, inpos, 0)) return 2;
   }
bufptr = 0;
return 0;
}
